"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = require("chalk");
const ora_1 = __importDefault(require("ora"));
const pluralize_1 = __importDefault(require("pluralize"));
const helpers_1 = require("yargs/helpers");
const yargs_1 = __importDefault(require("yargs/yargs"));
const config_1 = require("./config");
const error_1 = require("./error");
const events_1 = require("./events");
const smoker_1 = require("./smoker");
const util_1 = require("./util");
const BEHAVIOR_GROUP = 'Behavior:';
/**
 * Main entry point for the CLI script
 * @param args Raw command-line arguments array
 */
async function main(args) {
    const y = (0, yargs_1.default)(args);
    const result = await (0, util_1.readPackageJson)({ cwd: __dirname });
    if (!result) {
        throw new Error('midnight-smoker could not find its own package.json! bogus');
    }
    const { packageJson } = result;
    const { version, homepage } = packageJson;
    /**
     * These options can be specified more than once.
     *
     * Unfortunately, the typing gets hinky if we try to define the options object outside of the `options()` call, so we cannot programmatically gather these.
     */
    const arrayOptNames = new Set([
        'add',
        'workspace',
        'pm',
        'script',
        'scripts',
    ]);
    try {
        const config = await (0, config_1.readConfig)();
        await y
            .scriptName('smoker')
            .command('* [scripts..]', 'Run tests against a package as it would be published', (yargs) => {
            /**
             * Reusable config for array-type options
             */
            const arrayOptConfig = {
                requiresArg: true,
                array: true,
                type: 'string',
                coerce: util_1.normalizeStringArray,
            };
            return yargs
                .positional('scripts', {
                describe: 'Script(s) in package.json to run',
                type: 'string',
                coerce: util_1.normalizeStringArray,
                default: config.scripts,
            })
                .options({
                add: {
                    describe: 'Additional dependency to provide to script(s)',
                    group: BEHAVIOR_GROUP,
                    default: config.add,
                    ...arrayOptConfig,
                },
                all: {
                    describe: 'Run script in all workspaces',
                    group: BEHAVIOR_GROUP,
                    default: config.all,
                    type: 'boolean',
                },
                bail: {
                    describe: 'When running scripts, halt on first error',
                    group: BEHAVIOR_GROUP,
                    default: config.bail,
                    type: 'boolean',
                },
                'include-root': {
                    describe: "Include the workspace root; must provide '--all'",
                    group: BEHAVIOR_GROUP,
                    default: config.includeRoot,
                    implies: 'all',
                    type: 'boolean',
                },
                json: {
                    describe: 'Output JSON only',
                    default: config.json,
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
                linger: {
                    describe: 'Do not clean up temp dir(s) after completion',
                    group: BEHAVIOR_GROUP,
                    default: config.linger,
                    hidden: true,
                    type: 'boolean',
                },
                pm: {
                    describe: 'Run script(s) with a specific package manager; <npm|yarn|pnpm>[@version]',
                    group: BEHAVIOR_GROUP,
                    default: config.pm ?? 'npm@latest',
                    ...arrayOptConfig,
                },
                loose: {
                    describe: 'Ignore missing scripts (used with --all)',
                    type: 'boolean',
                    default: config.loose,
                    group: BEHAVIOR_GROUP,
                    implies: 'all',
                },
                workspace: {
                    describe: 'Run script in a specific workspace or workspaces',
                    group: BEHAVIOR_GROUP,
                    default: config.workspace,
                    ...arrayOptConfig,
                },
                checks: {
                    describe: 'Run built-in checks',
                    group: BEHAVIOR_GROUP,
                    default: Boolean('checks' in config ? config.checks : true),
                    type: 'boolean',
                },
            })
                .check((argv) => {
                if (argv.pm?.some((pm) => pm.startsWith('pnpm'))) {
                    throw new error_1.SmokerError('pnpm is currently unsupported');
                }
                return true;
            });
        }, async (argv) => {
            const scripts = argv.scripts ?? [];
            let smoker;
            try {
                smoker = await smoker_1.Smoker.init(scripts, argv);
            }
            catch (err) {
                console.error((0, chalk_1.red)(err.message));
                process.exitCode = 1;
                return;
            }
            if (argv.json) {
                let totalPackages = null;
                let totalScripts = null;
                let totalPackageManagers = null;
                let failedScripts = null;
                let passedScripts = null;
                let totalChecks = null;
                let failedChecks = null;
                let passedChecks = null;
                smoker
                    .once(events_1.Events.INSTALL_BEGIN, ({ uniquePkgs, packageManagers }) => {
                    totalPackages = uniquePkgs.length;
                    totalPackageManagers = packageManagers.length;
                })
                    .once(events_1.Events.RUN_SCRIPTS_BEGIN, ({ total }) => {
                    totalScripts = total;
                })
                    .once(events_1.Events.RUN_SCRIPTS_FAILED, ({ failed, passed }) => {
                    failedScripts = failed;
                    passedScripts = passed;
                    smoker.removeAllListeners(events_1.Events.RUN_SCRIPTS_OK);
                    process.exitCode = 1;
                })
                    .once(events_1.Events.RUN_SCRIPTS_OK, ({ passed }) => {
                    passedScripts = passed;
                    failedScripts = 0;
                    smoker.removeAllListeners(events_1.Events.RUN_SCRIPTS_FAILED);
                })
                    .once(events_1.Events.RUN_CHECKS_BEGIN, ({ total }) => {
                    totalChecks = total;
                })
                    .once(events_1.Events.RUN_CHECKS_FAILED, ({ failed, passed }) => {
                    failedChecks = failed.length;
                    passedChecks = passed.length;
                    smoker.removeAllListeners(events_1.Events.RUN_CHECKS_OK);
                    process.exitCode = 1;
                })
                    .once(events_1.Events.RUN_CHECKS_OK, ({ passed }) => {
                    passedChecks = passed.length;
                    failedChecks = 0;
                    smoker.removeAllListeners(events_1.Events.RUN_CHECKS_FAILED);
                })
                    .once(events_1.Events.SMOKE_FAILED, () => {
                    process.exitCode = 1;
                });
                let output;
                try {
                    const results = await smoker.smoke();
                    const stats = {
                        totalPackages,
                        totalPackageManagers,
                        totalScripts,
                        failedScripts,
                        passedScripts,
                        totalChecks,
                        failedChecks,
                        passedChecks,
                    };
                    output = { results, stats };
                }
                catch (err) {
                    const stats = {
                        totalPackages,
                        totalPackageManagers,
                        totalScripts,
                        failedScripts,
                        passedScripts,
                        totalChecks,
                        failedChecks,
                        passedChecks,
                    };
                    output = { error: err.message, stats };
                }
                finally {
                    smoker.removeAllListeners();
                }
                console.log(JSON.stringify(output, null, 2));
            }
            else {
                const spinner = (0, ora_1.default)();
                const scriptFailedEvts = [];
                const checkFailedEvts = [];
                smoker
                    .on(events_1.Events.SMOKE_BEGIN, () => {
                    console.error(`💨 ${(0, chalk_1.blue)('midnight-smoker')} ${(0, chalk_1.white)(`v${version}`)}`);
                })
                    .on(events_1.Events.PACK_BEGIN, () => {
                    let what;
                    if (argv.workspace?.length) {
                        what = (0, pluralize_1.default)('workspace', argv.workspace.length, true);
                    }
                    else if (argv.all) {
                        what = 'all workspaces';
                        if (argv.includeRoot) {
                            what += ' (and the workspace root)';
                        }
                    }
                    else {
                        what = 'current project';
                    }
                    spinner.start(`Packing ${what}…`);
                })
                    .on(events_1.Events.PACK_OK, ({ uniquePkgs, packageManagers }) => {
                    let msg = `Packed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} using `;
                    if (packageManagers.length > 1) {
                        msg += `${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += `${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.succeed(msg);
                })
                    .on(events_1.Events.PACK_FAILED, (err) => {
                    spinner.fail(err.message);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.INSTALL_BEGIN, ({ uniquePkgs, packageManagers, additionalDeps }) => {
                    let msg = `Installing ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`;
                    if (additionalDeps.length) {
                        msg += ` with ${(0, pluralize_1.default)('additional dependency', additionalDeps.length, true)}`;
                    }
                    if (packageManagers.length > 1) {
                        msg += ` using ${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += ` using ${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.start(msg);
                })
                    .on(events_1.Events.INSTALL_FAILED, (err) => {
                    spinner.fail(err.message);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.INSTALL_OK, ({ uniquePkgs }) => {
                    spinner.succeed(`Installed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`);
                })
                    .on(events_1.Events.RUN_CHECKS_BEGIN, ({ total }) => {
                    spinner.start(`Running 0/${total} checks…`);
                })
                    .on(events_1.Events.RUN_CHECK_BEGIN, ({ current, total }) => {
                    spinner.text = `Running check ${current}/${total}…`;
                })
                    .on(events_1.Events.RUN_CHECK_FAILED, (evt) => {
                    checkFailedEvts.push(evt);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.RUN_CHECKS_OK, ({ total }) => {
                    spinner.succeed(`Successfully ran ${(0, pluralize_1.default)('check', total, true)}`);
                })
                    .on(events_1.Events.RUN_CHECKS_FAILED, ({ total, failed }) => {
                    spinner.fail(`${(0, pluralize_1.default)('check', failed.length, true)} out of ${total} failed`);
                    // TODO: move this crunching somewhere else
                    const failedByPackage = checkFailedEvts
                        .map((evt) => evt.failed)
                        .flat()
                        .reduce((acc, failed) => {
                        const pkgName = failed.context.pkgJson.name ?? failed.context.pkgPath;
                        acc[pkgName] = [...(acc[pkgName] ?? []), failed];
                        return acc;
                    }, {});
                    for (const [pkgName, failed] of Object.entries(failedByPackage)) {
                        let text = `${(0, chalk_1.red)('Check failure')} details for package ${(0, chalk_1.cyan)(pkgName)}:\n`;
                        for (const failure of failed) {
                            text += `» ${(0, chalk_1.yellow)(failure.message)}\n`;
                        }
                        spinner.info(text);
                    }
                    process.exitCode = 1;
                })
                    .on(events_1.Events.RUN_SCRIPTS_BEGIN, ({ total }) => {
                    spinner.start(`Running script 0/${total}…`);
                })
                    .on(events_1.Events.RUN_SCRIPT_BEGIN, ({ current, total }) => {
                    spinner.text = `Running script ${current}/${total}…`;
                })
                    .on(events_1.Events.RUN_SCRIPT_FAILED, (evt) => {
                    scriptFailedEvts.push(evt);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.RUN_SCRIPTS_OK, ({ total }) => {
                    spinner.succeed(`Successfully ran ${(0, pluralize_1.default)('script', total, true)}`);
                })
                    .on(events_1.Events.RUN_SCRIPTS_FAILED, ({ total, failed: failures }) => {
                    spinner.fail(`${failures} of ${total} ${(0, pluralize_1.default)('script', total)} failed`);
                    for (const evt of scriptFailedEvts) {
                        spinner.info(`${(0, chalk_1.red)('Script failure')} details for package ${(0, chalk_1.cyan)(evt.pkgName)}:\n» ${(0, chalk_1.yellow)(evt.error.message)}\n`);
                    }
                    process.exitCode = 1;
                })
                    .on(events_1.Events.SMOKE_FAILED, (err) => {
                    spinner.fail(err?.message ?? err);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.SMOKE_OK, () => {
                    spinner.succeed('Lovey-dovey! 💖');
                })
                    .on(events_1.Events.LINGERED, (dirs) => {
                    console.error(`Lingering ${(0, pluralize_1.default)('temp directory', dirs.length)}:\n`);
                    for (const dir of dirs) {
                        console.error(`  ${(0, chalk_1.yellow)(dir)}`);
                    }
                });
                await smoker.smoke();
            }
        })
            .options({
            verbose: {
                describe: 'Verbose output',
                type: 'boolean',
                global: true,
            },
        })
            .epilog(`For more info, see ${homepage}\n`)
            .middleware(
        /**
         * If an array-type option is provided in both the config file and on the command-line,
         * we use this to merge the two arrays.
         *
         * If we had any object-type options (we don't) then we'd do that here as well.
         */
        (argv) => {
            for (const key of arrayOptNames) {
                const cfgKey = key;
                const arg = key;
                if (cfgKey in config && arg in argv) {
                    const cfgValue = (Array.isArray(config[cfgKey])
                        ? config[cfgKey]
                        : [config[cfgKey]]);
                    const argvValue = (Array.isArray(argv[arg]) ? argv[arg] : [argv[arg]]);
                    argv[arg] = [...new Set([...cfgValue, ...argvValue])];
                }
            }
            // squelch some output if `json` is true
            argv.verbose = argv.json ? false : argv.verbose;
        })
            .showHelpOnFail(false)
            .help()
            .strict()
            .parseAsync();
    }
    catch (err) {
        console.error((0, chalk_1.red)(err));
        process.exitCode = 1;
    }
}
main((0, helpers_1.hideBin)(process.argv));
//# sourceMappingURL=cli.js.map