"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.readConfig = exports.SmokerConfigSchema = void 0;
const zod_1 = require("zod");
const debug_1 = __importDefault(require("debug"));
const lilconfig_1 = require("lilconfig");
const util_1 = require("./util");
const rules_1 = require("./rules");
const debug = (0, debug_1.default)('midnight-smoker:config');
async function loadEsm(filepath) {
    return import(filepath);
}
const DEFAULT_OPTS = Object.freeze({
    loaders: { '.mjs': loadEsm, '.js': loadEsm },
    searchPlaces: [
        'package.json',
        '.smokerrc.json',
        '.smokerrc.js',
        '.smokerrc.cjs',
        '.smokerrc.mjs',
        'smoker.config.json',
        'smoker.config.js',
        'smoker.config.cjs',
        'smoker.config.mjs',
        '.config/smokerrc.json',
        '.config/smokerrc.js',
        '.config/smokerrc.cjs',
        '.config/smokerrc.mjs',
        '.config/smoker.config.json',
        '.config/smoker.config.js',
        '.config/smoker.config.cjs',
        '.config/smoker.config.mjs',
    ],
});
exports.SmokerConfigSchema = zod_1.z
    .object({
    add: zod_1.z
        .union([zod_1.z.string(), zod_1.z.array(zod_1.z.string())])
        .optional()
        .describe('Add an extra package to the list of packages to be installed.'),
    all: zod_1.z
        .boolean()
        .optional()
        .describe('Operate on all workspaces. The root workspace is omitted unless `includeRoot` is `true`.'),
    bail: zod_1.z.boolean().optional().describe('Fail on first script failure.'),
    includeRoot: zod_1.z
        .boolean()
        .optional()
        .describe('Operate on the root workspace. Only has an effect if `all` is `true`.'),
    json: zod_1.z.boolean().optional().describe('Output JSON only.'),
    linger: zod_1.z
        .boolean()
        .optional()
        .describe('Do not delete temp directories after completion.'),
    verbose: zod_1.z.boolean().optional().describe('Verbose logging.'),
    workspace: zod_1.z
        .union([zod_1.z.string(), zod_1.z.array(zod_1.z.string())])
        .optional()
        .describe('One or more workspaces to run scripts in.'),
    pm: zod_1.z
        .union([zod_1.z.string(), zod_1.z.array(zod_1.z.string())])
        .optional()
        .describe('Package manager(s) to use.'),
    script: zod_1.z
        .union([zod_1.z.string(), zod_1.z.array(zod_1.z.string())])
        .optional()
        .describe('Script(s) to run. Alias of `scripts`.'),
    scripts: zod_1.z
        .union([zod_1.z.string(), zod_1.z.array(zod_1.z.string())])
        .optional()
        .describe('Script(s) to run. Alias of `script`.'),
    loose: zod_1.z
        .boolean()
        .optional()
        .describe('If `true`, fail if a workspace is missing a script.'),
    checks: zod_1.z
        .boolean()
        .optional()
        .describe('If `false`, run no builtin checks.'),
    rules: rules_1.RuleConfigSchema.optional().describe('Rule configuration for checks'),
})
    .describe('midnight-smoker configuration file schema');
const ARRAY_KEYS = new Set(['add', 'workspace', 'pm', 'script', 'scripts']);
async function readConfig() {
    const result = await (0, lilconfig_1.lilconfig)('smoker', DEFAULT_OPTS).search();
    if (result?.config && !result.config.isEmpty) {
        debug('Found config at %s', result.filepath);
        let config = result.config;
        // I love ESM, really I do
        if ('default' in config) {
            config = config.default;
        }
        // TODO actually use zod to validate this.
        // TODO also use transform or coerce or w/e to cast to array
        for (const key of ARRAY_KEYS) {
            const value = config[key];
            if (value) {
                config[key] = (0, util_1.castArray)(value);
            }
        }
        config.scripts = [...(config.script ?? []), ...(config.scripts ?? [])];
        delete config.script;
        debug('Loaded config: %O', config);
        return config;
    }
    return {};
}
exports.readConfig = readConfig;
//# sourceMappingURL=config.js.map