"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Npm7 = void 0;
const debug_1 = __importDefault(require("debug"));
const node_path_1 = __importDefault(require("node:path"));
const error_1 = require("../error");
const npm_1 = require("./npm");
class Npm7 extends npm_1.GenericNpmPackageManager {
    constructor(executor, opts = {}) {
        super(executor, opts);
        this.name = 'npm';
        this.debug = (0, debug_1.default)(`midnight-smoker:pm:npm7`);
    }
    static accepts(semver) {
        return Boolean(semver.compare('7.0.0') - semver.compare('9.0.0'));
    }
    static load(executor, opts) {
        return new Npm7(executor, opts);
    }
    async install(manifest) {
        const { packedPkgs, tarballRootDir } = manifest;
        if (!packedPkgs?.length) {
            throw new TypeError('(install) Non-empty "manifest" arg is required');
        }
        const additionalDeps = manifest.additionalDeps ?? [];
        // otherwise we get a deprecation warning
        const installArgs = [
            'install',
            '--no-package-lock',
            '--global-style',
            ...packedPkgs.map(({ tarballFilepath }) => tarballFilepath),
            ...additionalDeps,
        ];
        let installResult;
        try {
            installResult = await this.executor.exec(installArgs, {
                cwd: tarballRootDir,
            });
        }
        catch (err) {
            throw new error_1.SmokerError(`(install) ${this.name} failed to install: ${err.message}`);
        }
        if (installResult.exitCode) {
            this.debug('(install) Failed: %O', installResult);
            throw new error_1.SmokerError(`(install) Installation failed with exit code ${installResult.exitCode}: ${installResult.stderr}`);
        }
        this.debug('(install) Installed %d packages', packedPkgs.length);
        return installResult;
    }
    async pack(dest, opts = {}) {
        if (!dest) {
            throw new TypeError('(pack) "dest" arg is required');
        }
        let packArgs = [
            'pack',
            '--json',
            `--pack-destination=${dest}`,
            '--foreground-scripts=false', // suppress output of lifecycle scripts so json can be parsed
        ];
        if (opts.workspaces?.length) {
            packArgs = [
                ...packArgs,
                ...opts.workspaces.map((w) => `--workspace=${w}`),
            ];
        }
        else if (opts.allWorkspaces) {
            packArgs = [...packArgs, '--workspaces'];
            if (opts.includeWorkspaceRoot) {
                packArgs = [...packArgs, '--include-workspace-root'];
            }
        }
        let packResult;
        try {
            packResult = await this.executor.exec(packArgs);
        }
        catch (err) {
            throw new error_1.SmokerError(`(pack) ${this.name} failed to spawn: ${err.message}`);
        }
        if (packResult.exitCode) {
            this.debug('(pack) Failed: %O', packResult);
            throw new error_1.SmokerError(`(pack) Packing failed with exit code ${packResult.exitCode}: ${packResult.stderr}`);
        }
        let parsed;
        const { stdout: packOutput } = packResult;
        try {
            parsed = JSON.parse(packOutput);
        }
        catch {
            this.debug('(pack) Failed to parse JSON: %s', packOutput);
            throw new error_1.SmokerError(`(pack) Failed to parse JSON output from "${this.name} pack": ${packOutput}`);
        }
        const packedPkgs = parsed.map(({ filename, name }) => {
            // workaround for https://github.com/npm/cli/issues/3405
            filename = filename.replace(/^@(.+?)\//, '$1-');
            return {
                tarballFilepath: node_path_1.default.join(dest, filename),
                installPath: node_path_1.default.join(dest, 'node_modules', name),
                pkgName: name,
            };
        });
        this.debug('(pack) Packed %d packages', packedPkgs.length);
        return { packedPkgs, tarballRootDir: dest };
    }
}
exports.Npm7 = Npm7;
Npm7.bin = 'npm';
exports.default = Npm7;
//# sourceMappingURL=npm7.js.map