"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const promises_1 = __importDefault(require("node:fs/promises"));
const node_path_1 = __importDefault(require("node:path"));
const zod_1 = require("zod");
const rule_1 = require("../rule");
const debug = (0, debug_1.default)('midnight-smoker:rule:no-missing-pkg-files');
const noMissingPkgFiles = (0, rule_1.createRule)({
    async check({ pkgJson: pkg, pkgPath, fail }, opts) {
        const fieldsToCheck = opts?.fields ?? [];
        if (opts?.bin !== false) {
            fieldsToCheck.push('bin');
        }
        /**
         * @param relativePath Path from package root to file
         * @param field Field name
         * @param name Optional; refers to key in an object field (e.g., `bin` as an object instead of `string`)
         * @todo maybe make this a utility function?
         */
        const checkFile = async (relativePath, field, name) => {
            const filepath = node_path_1.default.resolve(pkgPath, relativePath);
            debug('Checking for %s from field %s', filepath, field);
            try {
                await promises_1.default.stat(filepath);
            }
            catch {
                return fail(name
                    ? `File "${name}" from "${field}" field unreadable at path: ${relativePath}`
                    : `File from "${field}" field unreadable at path: ${relativePath}`);
            }
        };
        const res = await Promise.all(fieldsToCheck
            .filter((field) => field in pkg)
            .map(async (field) => {
            const value = pkg[field];
            if (value) {
                if (typeof value === 'string') {
                    return checkFile(value, field);
                }
                else if (!Array.isArray(value) && typeof value === 'object') {
                    return Promise.all(Object.entries(value).map(([name, relativePath]) => {
                        return checkFile(relativePath, field, name);
                    }));
                }
            }
        }));
        return res.flat().filter(Boolean);
    },
    name: 'no-missing-pkg-files',
    description: 'Checks that files referenced in package.json exist in the tarball',
    schema: zod_1.z.object({
        bin: zod_1.z
            .boolean()
            .default(true)
            .optional()
            .describe('Check the "bin" field (if it exists)'),
        fields: zod_1.z
            .array(zod_1.z.string().min(1))
            .optional()
            .describe('Check files referenced by these additional top-level fields'),
    }),
});
exports.default = noMissingPkgFiles;
//# sourceMappingURL=no-missing-pkg-files.js.map